
package org.mule.modules.rest.processors;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.DefaultMuleEvent;
import org.mule.DefaultMuleMessage;
import org.mule.MessageExchangePattern;
import org.mule.RequestContext;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.MuleSession;
import org.mule.api.callback.SourceCallback;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.InterceptingMessageProcessor;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.modules.rest.RestModule;
import org.mule.modules.rest.adapters.RestModuleLifecycleAdapter;
import org.mule.modules.rest.config.NestedProcessorChain;
import org.mule.modules.rest.config.NestedProcessorString;
import org.mule.session.DefaultMuleSession;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * RouterMessageProcessor invokes the {@link org.mule.modules.rest.RestModule#router(java.lang.String, org.mule.api.NestedProcessor, org.mule.api.NestedProcessor, org.mule.api.NestedProcessor, org.mule.api.NestedProcessor, org.mule.api.NestedProcessor, java.lang.String, java.lang.String, org.mule.api.callback.SourceCallback)} method in {@link RestModule }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class RouterMessageProcessor
    implements SourceCallback, FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, InterceptingMessageProcessor
{

    private Object templateUri;
    private String _templateUriType;
    private Object get;
    private org.mule.api.NestedProcessor _getType;
    private Object put;
    private org.mule.api.NestedProcessor _putType;
    private Object post;
    private org.mule.api.NestedProcessor _postType;
    private Object delete;
    private org.mule.api.NestedProcessor _deleteType;
    private Object patch;
    private org.mule.api.NestedProcessor _patchType;
    private Object method;
    private String _methodType;
    private Object uri;
    private String _uriType;
    private static Logger logger = LoggerFactory.getLogger(RouterMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;
    /**
     * Message processor that will get called for processing incoming events
     * 
     */
    private MessageProcessor messageProcessor;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(RestModuleLifecycleAdapter.class);
                if (moduleObject == null) {
                    moduleObject = new RestModuleLifecycleAdapter();
                    muleContext.getRegistry().registerObject(RestModuleLifecycleAdapter.class.getName(), moduleObject);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.modules.rest.adapters.RestModuleLifecycleAdapter"), e, this);
            }
        }
        if (moduleObject instanceof String) {
            moduleObject = muleContext.getRegistry().lookupObject(((String) moduleObject));
            if (moduleObject == null) {
                throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object by config name"), this);
            }
        }
        if (post instanceof Initialisable) {
            ((Initialisable) post).initialise();
        }
        if (put instanceof Initialisable) {
            ((Initialisable) put).initialise();
        }
        if (get instanceof Initialisable) {
            ((Initialisable) get).initialise();
        }
        if (delete instanceof Initialisable) {
            ((Initialisable) delete).initialise();
        }
        if (patch instanceof Initialisable) {
            ((Initialisable) patch).initialise();
        }
    }

    public void start()
        throws MuleException
    {
        if (post instanceof Startable) {
            ((Startable) post).start();
        }
        if (put instanceof Startable) {
            ((Startable) put).start();
        }
        if (get instanceof Startable) {
            ((Startable) get).start();
        }
        if (delete instanceof Startable) {
            ((Startable) delete).start();
        }
        if (patch instanceof Startable) {
            ((Startable) patch).start();
        }
    }

    public void stop()
        throws MuleException
    {
        if (post instanceof Stoppable) {
            ((Stoppable) post).stop();
        }
        if (put instanceof Stoppable) {
            ((Stoppable) put).stop();
        }
        if (get instanceof Stoppable) {
            ((Stoppable) get).stop();
        }
        if (delete instanceof Stoppable) {
            ((Stoppable) delete).stop();
        }
        if (patch instanceof Stoppable) {
            ((Stoppable) patch).stop();
        }
    }

    public void dispose() {
        if (post instanceof Disposable) {
            ((Disposable) post).dispose();
        }
        if (put instanceof Disposable) {
            ((Disposable) put).dispose();
        }
        if (get instanceof Disposable) {
            ((Disposable) get).dispose();
        }
        if (delete instanceof Disposable) {
            ((Disposable) delete).dispose();
        }
        if (patch instanceof Disposable) {
            ((Disposable) patch).dispose();
        }
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
        if (post instanceof MuleContextAware) {
            ((MuleContextAware) post).setMuleContext(context);
        }
        if (put instanceof MuleContextAware) {
            ((MuleContextAware) put).setMuleContext(context);
        }
        if (get instanceof MuleContextAware) {
            ((MuleContextAware) get).setMuleContext(context);
        }
        if (delete instanceof MuleContextAware) {
            ((MuleContextAware) delete).setMuleContext(context);
        }
        if (patch instanceof MuleContextAware) {
            ((MuleContextAware) patch).setMuleContext(context);
        }
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
        if (post instanceof FlowConstructAware) {
            ((FlowConstructAware) post).setFlowConstruct(flowConstruct);
        }
        if (put instanceof FlowConstructAware) {
            ((FlowConstructAware) put).setFlowConstruct(flowConstruct);
        }
        if (get instanceof FlowConstructAware) {
            ((FlowConstructAware) get).setFlowConstruct(flowConstruct);
        }
        if (delete instanceof FlowConstructAware) {
            ((FlowConstructAware) delete).setFlowConstruct(flowConstruct);
        }
        if (patch instanceof FlowConstructAware) {
            ((FlowConstructAware) patch).setFlowConstruct(flowConstruct);
        }
    }

    /**
     * Sets the message processor that will "listen" the events generated by this message source
     * 
     * @param listener Message processor
     */
    public void setListener(MessageProcessor listener) {
        this.messageProcessor = listener;
    }

    /**
     * Implements {@link SourceCallback#process(org.mule.api.MuleEvent)}. This message source will be passed on to the actual pojo's method as a callback mechanism.
     * 
     */
    public Object process(Object message)
        throws Exception
    {
        MuleMessage muleMessage;
        muleMessage = new DefaultMuleMessage(message, muleContext);
        MuleSession muleSession;
        muleSession = new DefaultMuleSession(flowConstruct, muleContext);
        MuleEvent muleEvent;
        muleEvent = new DefaultMuleEvent(muleMessage, MessageExchangePattern.ONE_WAY, muleSession);
        try {
            MuleEvent responseEvent;
            responseEvent = messageProcessor.process(muleEvent);
            if ((responseEvent!= null)&&(responseEvent.getMessage()!= null)) {
                return responseEvent.getMessage().getPayload();
            }
        } catch (Exception e) {
            throw e;
        }
        return null;
    }

    /**
     * Implements {@link SourceCallback#process(org.mule.api.MuleEvent)}. This message source will be passed on to the actual pojo's method as a callback mechanism.
     * 
     */
    public Object process(Object message, Map<String, Object> properties)
        throws Exception
    {
        MuleMessage muleMessage;
        muleMessage = new DefaultMuleMessage(message, properties, null, null, muleContext);
        MuleSession muleSession;
        muleSession = new DefaultMuleSession(flowConstruct, muleContext);
        MuleEvent muleEvent;
        muleEvent = new DefaultMuleEvent(muleMessage, MessageExchangePattern.ONE_WAY, muleSession);
        try {
            MuleEvent responseEvent;
            responseEvent = messageProcessor.process(muleEvent);
            if ((responseEvent!= null)&&(responseEvent.getMessage()!= null)) {
                return responseEvent.getMessage().getPayload();
            }
        } catch (Exception e) {
            throw e;
        }
        return null;
    }

    /**
     * Implements {@link SourceCallback#process()}. This message source will be passed on to the actual pojo's method as a callback mechanism.
     * 
     */
    public Object process()
        throws Exception
    {
        try {
            MuleEvent responseEvent;
            responseEvent = messageProcessor.process(RequestContext.getEvent());
            if ((responseEvent!= null)&&(responseEvent.getMessage()!= null)) {
                return responseEvent.getMessage().getPayload();
            }
        } catch (Exception e) {
            throw e;
        }
        return null;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets post
     * 
     * @param value Value to set
     */
    public void setPost(Object value) {
        this.post = value;
    }

    /**
     * Sets put
     * 
     * @param value Value to set
     */
    public void setPut(Object value) {
        this.put = value;
    }

    /**
     * Sets get
     * 
     * @param value Value to set
     */
    public void setGet(Object value) {
        this.get = value;
    }

    /**
     * Sets templateUri
     * 
     * @param value Value to set
     */
    public void setTemplateUri(Object value) {
        this.templateUri = value;
    }

    /**
     * Sets delete
     * 
     * @param value Value to set
     */
    public void setDelete(Object value) {
        this.delete = value;
    }

    /**
     * Sets method
     * 
     * @param value Value to set
     */
    public void setMethod(Object value) {
        this.method = value;
    }

    /**
     * Sets uri
     * 
     * @param value Value to set
     */
    public void setUri(Object value) {
        this.uri = value;
    }

    /**
     * Sets patch
     * 
     * @param value Value to set
     */
    public void setPatch(Object value) {
        this.patch = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            if (((Class) expectedType).isPrimitive()) {
                if (((Class) expectedType).getName().equals("boolean")&&(clazz == Boolean.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("byte")&&(clazz == Byte.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("short")&&(clazz == Short.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("char")&&(clazz == Character.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("int")&&(clazz == Integer.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("float")&&(clazz == Float.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("long")&&(clazz == Long.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("double")&&(clazz == Double.class)) {
                    return true;
                }
                return false;
            } else {
                return ((Class) expectedType).isAssignableFrom(clazz);
            }
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, String expectedMimeType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, expectedMimeType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, expectedMimeType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, expectedMimeType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = null;
            if (expectedMimeType!= null) {
                targetDataType = DataTypeFactory.create(((Class) expectedType), expectedMimeType);
            } else {
                targetDataType = DataTypeFactory.create(((Class) expectedType));
            }
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage _muleMessage = event.getMessage();
        RestModuleLifecycleAdapter _castedModuleObject = null;
        if (moduleObject instanceof String) {
            _castedModuleObject = ((RestModuleLifecycleAdapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (_castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("router"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            _castedModuleObject = ((RestModuleLifecycleAdapter) moduleObject);
        }
        try {
            String _transformedTemplateUri = ((String) evaluateAndTransform(_muleMessage, RouterMessageProcessor.class.getDeclaredField("_templateUriType").getGenericType(), null, templateUri));
            org.mule.api.NestedProcessor _transformedGet = null;
            if ((get!= null)&&(get instanceof MessageProcessor)) {
                _transformedGet = new NestedProcessorChain(event, muleContext, ((MessageProcessor) get));
            } else {
                if ((get!= null)&&(get instanceof String)) {
                    _transformedGet = new NestedProcessorString(((String) get));
                }
            }
            org.mule.api.NestedProcessor _transformedPut = null;
            if ((put!= null)&&(put instanceof MessageProcessor)) {
                _transformedPut = new NestedProcessorChain(event, muleContext, ((MessageProcessor) put));
            } else {
                if ((put!= null)&&(put instanceof String)) {
                    _transformedPut = new NestedProcessorString(((String) put));
                }
            }
            org.mule.api.NestedProcessor _transformedPost = null;
            if ((post!= null)&&(post instanceof MessageProcessor)) {
                _transformedPost = new NestedProcessorChain(event, muleContext, ((MessageProcessor) post));
            } else {
                if ((post!= null)&&(post instanceof String)) {
                    _transformedPost = new NestedProcessorString(((String) post));
                }
            }
            org.mule.api.NestedProcessor _transformedDelete = null;
            if ((delete!= null)&&(delete instanceof MessageProcessor)) {
                _transformedDelete = new NestedProcessorChain(event, muleContext, ((MessageProcessor) delete));
            } else {
                if ((delete!= null)&&(delete instanceof String)) {
                    _transformedDelete = new NestedProcessorString(((String) delete));
                }
            }
            org.mule.api.NestedProcessor _transformedPatch = null;
            if ((patch!= null)&&(patch instanceof MessageProcessor)) {
                _transformedPatch = new NestedProcessorChain(event, muleContext, ((MessageProcessor) patch));
            } else {
                if ((patch!= null)&&(patch instanceof String)) {
                    _transformedPatch = new NestedProcessorString(((String) patch));
                }
            }
            String _transformedMethod = ((String) evaluateAndTransform(_muleMessage, RouterMessageProcessor.class.getDeclaredField("_methodType").getGenericType(), null, "#[header:INBOUND:http.method]"));
            String _transformedUri = ((String) evaluateAndTransform(_muleMessage, RouterMessageProcessor.class.getDeclaredField("_uriType").getGenericType(), null, "#[header:INBOUND:http.request.path]"));
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = _castedModuleObject.router(_transformedTemplateUri, _transformedGet, _transformedPut, _transformedPost, _transformedDelete, _transformedPatch, _transformedMethod, _transformedUri, this);
            TransformerTemplate.OverwitePayloadCallback overwritePayloadCallback = null;
            if (resultPayload == null) {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(NullPayload.getInstance());
            } else {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(resultPayload);
            }
            List<Transformer> transformerList;
            transformerList = new ArrayList<Transformer>();
            transformerList.add(new TransformerTemplate(overwritePayloadCallback));
            event.getMessage().applyTransformers(event, transformerList);
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("router"), event, e);
        }
    }

}
